from typing import Optional, Tuple

import torch
import torch.nn as nn

from velora.models.lnn.cell import NCPLiquidCell
from velora.models.lnn.sparse import SparseLinear
from velora.models.weight import WeightInitType
from velora.utils.torch import active_parameters, total_parameters
from velora.wiring import Wiring


class LiquidNCPNetwork(nn.Module):
    """
    A CfC Liquid Neural Circuit Policy (NCP) Network with three layers:

    1. Inter (input) - a `SparseLinear` layer
    2. Command (hidden) - a `NCPLiquidCell` layer
    3. Motor (output) - a `SparseLinear` layer

    ??? note "Decision nodes"

        `inter` and `command` neurons are automatically calculated using:

        ```python
        command_neurons = max(int(0.4 * n_neurons), 1)
        inter_neurons = n_neurons - command_neurons
        ```

    Combines a Liquid Time-Constant (LTC) cell with Ordinary Neural Circuits (ONCs). Paper references:

    - [Closed-form Continuous-time Neural Models](https://arxiv.org/abs/2106.13898)
    - [Reinforcement Learning with Ordinary Neural Circuits](https://proceedings.mlr.press/v119/hasani20a.html)
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        sparsity_level: float = 0.5,
        init_type: str | WeightInitType = "kaiming_uniform",
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            in_features (int): number of inputs (sensory nodes)
            n_neurons (int): number of decision nodes (inter and command nodes)
            out_features (int): number of out features (motor nodes)
            sparsity_level (float, optional): controls the connection sparsity
                between neurons.

                Must be a value between `[0.1, 0.9]` -

                - When `0.1` neurons are very dense.
                - When `0.9` they are very sparse.

            init_type (str, optional): the type of weight initialization
            device (torch.device, optional): the device to load tensors on
        """
        super().__init__()

        self.in_features = in_features
        self.n_neurons = n_neurons
        self.out_features = out_features
        self.device = device

        self.n_units = n_neurons + out_features  # inter + command + motor

        self.wiring = Wiring(
            in_features,
            n_neurons,
            out_features,
            sparsity_level=sparsity_level,
        )
        self.masks, self.counts = self.wiring.data()

        self.inter = SparseLinear(
            in_features,
            self.counts.inter,
            torch.abs(self.masks.inter.T),
            init_type=init_type,
            device=device,
        ).to(device)

        self.command = NCPLiquidCell(
            self.counts.inter,
            self.counts.command,
            self.masks.command,
            init_type=init_type,
            device=device,
        ).to(device)
        self.hidden_size = self.counts.command

        self.motor = SparseLinear(
            self.counts.command,
            self.counts.motor,
            torch.abs(self.masks.motor.T),
            init_type=init_type,
            device=device,
        ).to(device)

        self.act = nn.Mish()

        self._total_params = total_parameters(self)
        self._active_params = active_parameters(self)

    @property
    def total_params(self) -> int:
        """
        Gets the network's total parameter count.

        Returns:
            count (int): the total parameter count.
        """
        return self._total_params

    @property
    def active_params(self) -> int:
        """
        Gets the network's active parameter count.

        Returns:
            count (int): the active parameter count.
        """
        return self._active_params

    def forward(
        self, x: torch.Tensor, h_state: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            x (torch.Tensor): an input tensor of shape: `(batch_size, features)`.

                - `batch_size` the number of samples per timestep.
                - `features` the features at each timestep (e.g.,
                image features, joint coordinates, word embeddings, raw amplitude
                values).
            h_state (torch.Tensor, optional): initial hidden state of the RNN with
                shape: `(batch_size, n_units)`.

                - `batch_size` the number of samples.
                - `n_units` the total number of hidden neurons
                    (`n_neurons + out_features`).

        Returns:
            y_pred (torch.Tensor): the network prediction. When `batch_size=1`. Out shape is `(out_features)`. Otherwise, `(batch_size, out_features)`.
            h_state (torch.Tensor): the final hidden state. Output shape is `(batch_size, n_units)`.
        """
        if x.dim() != 2:
            raise ValueError(
                f"Unsupported dimensionality: '{x.shape}'. Should be 2 dimensional with: '(batch_size, features)'."
            )

        x = x.to(dtype=torch.float32, device=self.device)

        batch_size, features = x.size()

        if h_state is None:
            h_state = torch.zeros(
                (batch_size, self.hidden_size),
                device=self.device,
            )

        # Batch -> (batch_size, out_features)
        x = self.act(self.inter(x))
        x, h_state = self.command(x, h_state.to(self.device))
        y_pred: torch.Tensor = self.motor(self.act(x))

        # Single item -> (out_features)
        if y_pred.shape[0] == 1:
            y_pred = y_pred.squeeze(0)

        # h_state -> (batch_size, n_units)
        return y_pred, h_state


class NCPNetwork(nn.Module):
    """
    A Neural Circuit Policy (NCP) Network with three layers:

    1. Inter (input) - a `SparseLinear` layer
    2. Command (hidden) - a `SparseLinear` layer
    3. Motor (output) - a `SparseLinear` layer

    Uses the Mish activation function between each layer.

    ??? note "Decision nodes"

        `inter` and `command` neurons are automatically calculated using:

        ```python
        command_neurons = max(int(0.4 * n_neurons), 1)
        inter_neurons = n_neurons - command_neurons
        ```

    Uses an Ordinary Neural Circuit (ONC) architecture without Liquid dynamics.
    Paper references:

    - [Reinforcement Learning with Ordinary Neural Circuits](https://proceedings.mlr.press/v119/hasani20a.html)
    - [Mish: A Self Regularized Non-Monotonic Activation Function](https://arxiv.org/abs/1908.08681)
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        sparsity_level: float = 0.5,
        init_type: str | WeightInitType = "kaiming_uniform",
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            in_features (int): number of inputs (sensory nodes)
            n_neurons (int): number of decision nodes (inter and command nodes)
            out_features (int): number of out features (motor nodes)
            sparsity_level (float, optional): controls the connection sparsity
                between neurons.

                Must be a value between `[0.1, 0.9]` -

                - When `0.1` neurons are very dense.
                - When `0.9` they are very sparse.

            init_type (str, optional): the type of weight initialization
            device (torch.device, optional): the device to load tensors on
        """
        super().__init__()

        self.in_features = in_features
        self.n_neurons = n_neurons
        self.out_features = out_features
        self.device = device

        self.n_units = n_neurons + out_features  # inter + command + motor

        self.wiring = Wiring(
            in_features,
            n_neurons,
            out_features,
            sparsity_level=sparsity_level,
        )
        self.masks, self.counts = self.wiring.data()

        self.ncp = nn.Sequential(
            SparseLinear(
                in_features,
                self.counts.inter,
                torch.abs(self.masks.inter.T),
                init_type=init_type,
                device=device,
            ),
            nn.Mish(),
            SparseLinear(
                self.counts.inter,
                self.counts.command,
                torch.abs(self.masks.command.T),
                init_type=init_type,
                device=device,
            ),
            nn.Mish(),
            SparseLinear(
                self.counts.command,
                self.counts.motor,
                torch.abs(self.masks.motor.T),
                init_type=init_type,
                device=device,
            ),
        ).to(device)

        self._total_params = total_parameters(self)
        self._active_params = active_parameters(self)

    @property
    def total_params(self) -> int:
        """
        Gets the network's total parameter count.

        Returns:
            count (int): the total parameter count.
        """
        return self._total_params

    @property
    def active_params(self) -> int:
        """
        Gets the network's active parameter count.

        Returns:
            count (int): the active parameter count.
        """
        return self._active_params

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Performs a forward pass through the network.

        Parameters:
            x (torch.Tensor): an input tensor of shape: `(batch_size, features)`.

                - `batch_size` the number of samples per timestep.
                - `features` the features at each timestep (e.g.,
                image features, joint coordinates, word embeddings, raw amplitude
                values).

        Returns:
            y_pred (torch.Tensor): the network prediction. When `batch_size=1`. Out shape is `(out_features)`. Otherwise, `(batch_size, out_features)`.
        """
        if x.dim() != 2:
            raise ValueError(
                f"Unsupported dimensionality: '{x.shape}'. Should be 2 dimensional with: '(batch_size, features)'."
            )

        x = x.to(dtype=torch.float32, device=self.device)

        # Batch -> (batch_size, out_features)
        y_pred: torch.Tensor = self.ncp(x)

        # Single item -> (out_features)
        if y_pred.shape[0] == 1:
            y_pred = y_pred.squeeze(0)

        return y_pred



from velora.training.handler import TrainHandler
from velora.training.metrics import MovingMetric, StepStorage, TrainMetrics

__all__ = [
    "StepStorage",
    "MovingMetric",
    "TrainMetrics",
    "TrainHandler",
]



from velora.models.lnn.cell import NCPLiquidCell
from velora.models.lnn.ncp import LiquidNCPNetwork, NCPNetwork
from velora.models.lnn.sparse import SparseLinear

__all__ = [
    "NCPLiquidCell",
    "LiquidNCPNetwork",
    "NCPNetwork",
    "SparseLinear",
]



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from velora.models.base import RLModuleAgent  # pragma: no cover
    from velora.callbacks import TrainCallback  # pragma: no cover

from velora.utils.format import number_to_short

NAME_STR = """
__     __   _
\\ \\   / /__| | ___  _ __ __ _
 \\ \\ / / _ \\ |/ _ \\| '__/ _` |
  \\ V /  __/ | (_) | | | (_| |
   \\_/ \\___|_|\\___/|_|  \\__,_|
"""


def training_info(
    agent: "RLModuleAgent",
    n_episodes: int,
    batch_size: int,
    window_size: int,
    warmup_steps: int,
    callbacks: List["TrainCallback"],
) -> None:
    """
    Display's starting information to the console for a training run.

    Parameters:
        agent (Any): the agent being trained
        n_episodes (int): maximum number of training episodes
        batch_size (int): sampling batch size
        window_size (int): moving average window size
        warmup_steps (int): number of buffer warmup steps
        callbacks (List[TrainCallback]): applied training callbacks
    """
    output = NAME_STR.strip()
    params_str = f"{agent.active_params:,}/{agent.total_params:,}"

    if agent.active_params > 10_000:
        active, total = (
            number_to_short(agent.active_params),
            number_to_short(agent.total_params),
        )
        params_str += f" ({active}/{total})"

    cb_str = "\n\nActive Callbacks:"
    cb_str += "\n---------------------------------\n"
    cb_str += "\n".join(cb.info().lstrip() for cb in callbacks)
    cb_str += "\n---------------------------------\n"

    output += cb_str if callbacks else "\n\n"
    output += f"Training '{agent.__class__.__name__}' agent on '{agent.env.spec.id}' for '{number_to_short(n_episodes)}' episodes.\n"
    output += f"Using '{agent.buffer.__class__.__name__}' with 'capacity={number_to_short(agent.buffer.capacity)}'.\n"
    output += f"Sampling episodes with '{batch_size=}'.\n"
    output += f"Warming buffer with '{number_to_short(warmup_steps)}' samples before training starts.\n"
    output += f"Generating random values with a seed of '{agent.seed}'.\n"
    output += f"Running computations on device '{agent.device}'.\n"
    output += f"Moving averages computed based on 'window_size={number_to_short(window_size)}'.\n"
    output += f"Using networks with '{params_str}' active parameters.\n"
    output += "---------------------------------"

    print(output)



import random

import numpy as np
import torch


def set_seed(value: int | None = None) -> int:
    """
    Sets the random seed for `Python`, `PyTorch` and `NumPy`.
    When `None` will create a new one automatically.

    Parameters:
        value (int, optional): the seed value

    Returns:
        seed (int): the used seed value
    """
    if value is None:
        value = random.randint(0, 2**32 - 1)

    random.seed(value)
    torch.manual_seed(value)
    np.random.seed(value)

    return value


def set_device(device: str = "auto") -> torch.device:
    """
    Sets the `PyTorch` device dynamically.

    Parameters:
        device (str, optional): the name of the device to perform computations on.

            When `auto`:

            - Set to `cuda:0`, if available.
            - Else, `cpu`.

    Returns:
        device (torch.device): the `PyTorch` device.
    """
    if device == "auto":
        device = "cuda:0" if torch.cuda.is_available() else "cpu"

    return torch.device(device)



import json
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, Literal, Self

import torch
from safetensors.torch import load_file, save_file

from velora.buffer.experience import BatchExperience

MetaDataKeys = Literal[
    "capacity",
    "state_dim",
    "action_dim",
    "hidden_dim",
    "position",
    "size",
    "device",
]
BufferKeys = Literal[
    "states",
    "actions",
    "rewards",
    "next_states",
    "dones",
    "hiddens",
]


class BufferBase:
    """
    A base class for all buffers.

    Stores experiences `(states, actions, rewards, next_states, dones)` as
    individual items in tensors.
    """

    def __init__(
        self,
        capacity: int,
        state_dim: int,
        action_dim: int,
        hidden_dim: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            capacity (int): the total capacity of the buffer
            state_dim (int): dimension of state observations
            action_dim (int): dimension of actions
            hidden_dim (int): dimension of hidden state
            device (torch.device, optional): the device to perform computations on
        """
        self.capacity = capacity
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.hidden_dim = hidden_dim
        self.device = device

        # Position indicators
        self.position = 0
        self.size = 0

        # Pre-allocate storage
        self.states = torch.zeros((capacity, state_dim), device=device)
        self.actions = torch.zeros((capacity, action_dim), device=device)
        self.rewards = torch.zeros((capacity, 1), device=device)
        self.next_states = torch.zeros((capacity, state_dim), device=device)
        self.dones = torch.zeros((capacity, 1), device=device)
        self.hiddens = torch.zeros((capacity, hidden_dim), device=device)

    def add(
        self,
        state: torch.Tensor,
        action: torch.Tensor,
        reward: float,
        next_state: torch.Tensor,
        done: bool,
        hidden: torch.Tensor,
    ) -> None:
        """
        Adds a single experience to the buffer.

        Parameters:
            state (torch.Tensor): current state observation
            action (torch.Tensor): action taken
            reward (float): reward received
            next_state (torch.Tensor): next state observation
            done (bool): whether the episode ended
            hidden (torch.Tensor): Actor hidden state (prediction network)
        """
        self.states[self.position] = state.to(torch.float32)
        self.actions[self.position] = action
        self.rewards[self.position] = reward
        self.next_states[self.position] = next_state.to(torch.float32)
        self.dones[self.position] = done
        self.hiddens[self.position] = hidden

        # Update position - deque style
        self.position = (self.position + 1) % self.capacity
        self.size = min(self.size + 1, self.capacity)

    def add_multi(
        self,
        states: torch.Tensor,
        actions: torch.Tensor,
        rewards: torch.Tensor,
        next_states: torch.Tensor,
        dones: torch.Tensor,
        hiddens: torch.Tensor,
    ) -> None:
        """
        Adds a set of experience to the buffer.

        Parameters:
            states (torch.Tensor): current state observations
            actions (torch.Tensor): action takens
            rewards (torch.Tensor): rewards received
            next_states (torch.Tensor): next state observations
            dones (torch.Tensor): whether the episode ended
            hiddens (torch.Tensor): Actor hidden states (prediction network)
        """
        batch_size = states.shape[0]

        new_position = self.position + batch_size
        indices = torch.arange(self.position, new_position) % self.capacity

        dtype = torch.float32
        rewards = rewards.unsqueeze(-1) if rewards.dim() == 1 else rewards
        dones = dones.unsqueeze(-1) if dones.dim() == 1 else dones
        actions = actions.unsqueeze(-1) if actions.dim() == 1 else actions

        self.states[indices] = states.to(dtype)
        self.actions[indices] = actions.to(dtype)
        self.rewards[indices] = rewards.to(dtype)
        self.next_states[indices] = next_states.to(dtype)
        self.dones[indices] = dones.to(dtype)
        self.hiddens[indices] = hiddens

        # Update position - deque style
        self.position = (new_position) % self.capacity
        self.size = min(self.size + batch_size, self.capacity)

    @abstractmethod
    def sample(self) -> BatchExperience:
        """
        Samples experience from the buffer.

        Returns:
            batch (BatchExperience): an object of samples with the attributes (`states`, `actions`, `rewards`, `next_states`, `dones`, `hidden`).

                All items have the same shape `(batch_size, features)`.
        """
        pass  # pragma: no cover

    def __len__(self) -> int:
        """
        Gets the current size of the buffer.

        Returns:
            size (int): the current size of the buffer.
        """
        return self.size

    def metadata(self) -> Dict[MetaDataKeys, Any]:
        """
        Gets the metadata of the buffer.

        Includes:

        - `capacity` - the maximum capacity of the buffer.
        - `state_dim` - state dimension.
        - `action_dim` - action dimension.
        - `hidden_dim` - hidden state dimension.
        - `position` - current buffer position.
        - `size` - current size of buffer.
        - `device` - the device used for computations.

        Returns:
            metadata (Dict[str, Any]): the buffers metadata
        """
        return {
            "capacity": self.capacity,
            "state_dim": self.state_dim,
            "action_dim": self.action_dim,
            "hidden_dim": self.hidden_dim,
            "position": self.position,
            "size": self.size,
            "device": str(self.device) if self.device else None,
        }

    def state_dict(self) -> Dict[BufferKeys, torch.Tensor]:
        """
        Return a dictionary containing the buffers state.

        Includes:

        - `states` - tensor of states.
        - `actions` - tensor of actions.
        - `rewards` - tensor of rewards.
        - `next_states` - tensor of next states.
        - `dones` - tensor of dones.
        - `hiddens` - tensor of Actor hidden states (prediction network).

        Returns:
            state_dict (Dict[str, torch.Tensor]): the current state of the buffer
        """
        return {
            "states": self.states,
            "actions": self.actions,
            "rewards": self.rewards,
            "next_states": self.next_states,
            "dones": self.dones,
            "hiddens": self.hiddens,
        }

    def save(self, dirpath: str | Path, prefix: str = "buffer_") -> None:
        """
        Saves a buffers `state_dict()` to a `safetensors` file.

        Includes:

        - `<prefix>metadata.json` - the buffers metadata
        - `<prefix>state.safetensors` - the buffer state

        Parameters:
            dirpath (str | Path): the folder path to save the buffer state
            prefix (str, optional): a name prefix for the files
        """
        save_path = Path(dirpath)
        save_path.mkdir(parents=True, exist_ok=True)

        metadata_path = Path(save_path, f"{prefix}metadata").with_suffix(".json")
        buffer_path = Path(save_path, f"{prefix}state").with_suffix(".safetensors")

        save_file(self.state_dict(), buffer_path)

        with metadata_path.open("w") as f:
            f.write(json.dumps(self.metadata(), indent=2))

    @classmethod
    def load(cls, state_path: str | Path, metadata: Dict[MetaDataKeys, Any]) -> Self:
        """
        Restores the buffer from a saved state.

        Parameters:
            state_path (str | Path): the filepath to the buffer state
            metadata (Dict[str, Any]): a dictionary of metadata already
                loaded from a `metadata.json` file

        Returns:
            buffer (Self): a new buffer instance with the saved state restored
        """
        buffer_path = Path(state_path).with_suffix(".safetensors")
        device = metadata["device"] or "cpu"

        # Create new buffer instance
        buffer = cls(
            capacity=metadata["capacity"],
            state_dim=metadata["state_dim"],
            action_dim=metadata["action_dim"],
            hidden_dim=metadata["hidden_dim"],
            device=torch.device(device) if device else None,
        )
        buffer.position = metadata["position"]
        buffer.size = metadata["size"]

        # Load buffer state
        data: Dict[BufferKeys, torch.Tensor] = load_file(buffer_path, device)

        buffer.states = data["states"]
        buffer.actions = data["actions"]
        buffer.rewards = data["rewards"]
        buffer.next_states = data["next_states"]
        buffer.dones = data["dones"]
        buffer.hiddens = data["hiddens"]

        return buffer



from velora.utils.capture import evaluate_agent, record_episode
from velora.utils.core import set_device, set_seed
from velora.utils.format import number_to_short
from velora.utils.restore import load_model, save_model
from velora.utils.torch import (
    active_parameters,
    hard_update,
    soft_update,
    stack_tensor,
    summary,
    to_tensor,
    total_parameters,
)

__all__ = [
    "set_device",
    "set_seed",
    "to_tensor",
    "stack_tensor",
    "soft_update",
    "hard_update",
    "total_parameters",
    "active_parameters",
    "summary",
    "record_episode",
    "number_to_short",
    "save_model",
    "load_model",
    "evaluate_agent",
]



from dataclasses import dataclass
from typing import Tuple

import numpy as np
import torch


@dataclass
class NeuronCounts:
    """
    Storage container for NCP neuron category counts.

    Parameters:
        sensory (int): number of input nodes
        inter (int): number of decision nodes
        command (int): number of high-level decision nodes
        motor (int): number of output nodes
    """

    sensory: int
    inter: int
    command: int
    motor: int


@dataclass
class SynapseCounts:
    """
    Storage container for NCP neuron synapse connection counts.

    Parameters:
        sensory (int): number of connections for input nodes
        inter (int): number of connections for decision nodes
        command (int): number of connections for high-level decision nodes
        motor (int): number of connections for output nodes
    """

    sensory: int
    inter: int
    command: int
    motor: int


@dataclass
class LayerMasks:
    """
    Storage container for layer masks.

    Parameters:
        inter (torch.Tensor): sparse weight mask for input layer
        command (torch.Tensor): sparse weight mask for hidden layer
        motor (torch.Tensor): sparse weight mask for output layer
        recurrent (torch.Tensor): sparse weight mask for recurrent connections
    """

    inter: torch.Tensor
    command: torch.Tensor
    motor: torch.Tensor
    recurrent: torch.Tensor


class Wiring:
    """
    Creates sparse wiring masks for Neural Circuit Policy (NCP) Networks.

    !!! note

        NCPs have three layers:

        1. Inter (input)
        2. Command (hidden)
        3. Motor (output)
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        sparsity_level: float = 0.5,
    ) -> None:
        """
        Parameters:
            in_features (int): number of inputs (sensory nodes)
            n_neurons (int): number of decision nodes (inter and command nodes)
            out_features (int): number of outputs (motor nodes)
            sparsity_level (float, optional): controls the connection sparsity between neurons.

                Must be a value between `[0.1, 0.9]` -

                - When `0.1` neurons are very dense.
                - When `0.9` neurons are very sparse.
        """
        if sparsity_level < 0.1 or sparsity_level > 0.9:
            raise ValueError(f"'{sparsity_level=}' must be between '[0.1, 0.9]'.")

        self.density_level = 1.0 - sparsity_level

        self.n_command = max(int(0.4 * n_neurons), 1)
        self.n_inter = n_neurons - self.n_command

        self.counts, self._n_connections = self._set_counts(
            in_features,
            out_features,
        )
        self.masks = self._init_masks(in_features)

        self.build()

    @property
    def n_connections(self) -> SynapseCounts:
        """
        Neuron connection counts.

        Returns:
            connections (SynapseCounts): object containing neuron connection counts.
        """
        return self._n_connections

    def _init_masks(self, n_inputs: int) -> LayerMasks:
        """
        Helper method. Initializes all layer masks with zeros
        and stores them in a container.

        Parameters:
            n_inputs (int): the number of input nodes in the layer

        Returns:
            masks (LayerMasks): initialized layer masks.
        """
        return LayerMasks(
            inter=torch.zeros(
                (n_inputs, self.counts.inter),
                dtype=torch.int32,
            ),
            command=torch.zeros(
                (self.counts.inter, self.counts.command),
                dtype=torch.int32,
            ),
            motor=torch.zeros(
                (self.counts.command, self.counts.motor),
                dtype=torch.int32,
            ),
            recurrent=torch.zeros(
                (self.counts.command, self.counts.command),
                dtype=torch.int32,
            ),
        )

    def _synapse_count(self, count: int, scale: int = 1) -> int:
        """
        Helper method. Computes the synapse count for a single layer.

        Parameters:
            count (int): the number of neurons
            scale (int, optional): a scale factor

        Returns:
            count (int): synapse count.
        """
        return max(int(count * self.density_level * scale), 1)

    def _set_counts(
        self, in_features: int, out_features: int
    ) -> Tuple[NeuronCounts, SynapseCounts]:
        """
        Helper method. Computes the node layer and connection counts.

        Parameters:
            in_features (int): number of network input nodes
            out_features (int): number of network output nodes

        Returns:
            neuron_counts (NeuronCounts): object with neuron counts.
            synapse_counts (SynapseCounts): object with synapse connection counts.
        """
        counts = NeuronCounts(
            sensory=in_features,
            inter=self.n_inter,
            command=self.n_command,
            motor=out_features,
        )

        connections = SynapseCounts(
            sensory=self._synapse_count(self.n_inter),
            inter=self._synapse_count(self.n_command),
            command=self._synapse_count(self.n_command, scale=2),
            motor=self._synapse_count(self.n_command),
        )

        return counts, connections

    @staticmethod
    def polarity(shape: Tuple[int, ...] = (1,)) -> torch.IntTensor:
        """
        Utility method. Randomly selects a polarity of `-1` or `1`, `n` times
        based on shape.

        Parameters:
            shape (Tuple[int, ...]): size of the polarity matrix to generate.

        Returns:
            matrix (torch.Tensor): a polarity matrix filled with `-1` and `1`.
        """
        return torch.IntTensor(np.random.choice([-1, 1], shape))

    def _build_connections(self, mask: torch.Tensor, count: int) -> torch.Tensor:
        """
        Helper method. Randomly assigns connections to a set of nodes by populating
        its mask.

        !!! note "Performs two operations"

            1. Applies minimum connections (count) to all nodes.
            2. Checks all nodes have at least 1 connection.
                If not, adds a connection to 'missing' nodes.

        Parameters:
            mask (torch.Tensor): the initialized mask
            count (int): the number of connections per node

        Examples:
            Given 2 sensory (input) nodes and 5 inter neurons, we can define
            our first layer (inter) mask as:

            ```python
            import torch

            inter_mask = torch.zeros((2, 5), dtype=torch.int32)
            n_connections = 2

            inter_mask = wiring._build_connections(inter_mask, n_connections)

            # tensor([[-1,  1,  0,  0,  1],
            #         [ 0,  0, -1, -1,  0]], dtype=torch.int32)
            ```

        Returns:
            mask (torch.Tensor): updated layer sparsity mask.
        """
        num_nodes, num_cols = mask.shape

        # Add required connection count
        col_indices = torch.IntTensor(
            np.random.choice(num_cols, (num_nodes, count)),
        )
        polarities = self.polarity(col_indices.shape)
        row_indices = torch.arange(num_nodes).unsqueeze(1)

        mask[row_indices, col_indices] = polarities

        # Add missing node connections (if applicable)
        # -> Every node in 'num_cols' must have at least 1 connection
        # -> Column with all 0s = non-connected node
        is_col_all_zero = (mask == 0).all(dim=0)
        col_zero_indices = torch.nonzero(is_col_all_zero, as_tuple=True)[0]
        zero_count = col_zero_indices.numel()

        if zero_count > 0:
            # For each missing connection, randomly select a node and add connection
            # -> row = node
            row_indices = torch.randint(0, num_nodes, (zero_count,))
            random_polarities = self.polarity((zero_count,))
            mask[row_indices, col_zero_indices] = random_polarities

        return mask

    def _build_recurrent_connections(
        self, array: torch.Tensor, count: int
    ) -> torch.Tensor:
        """
        Utility method. Adds recurrent connections to a set of nodes.

        Used to simulate bidirectional connections between command neurons. Strictly
        used for visualization purposes.

        Parameters:
            array (torch.Tensor): an initialized matrix to update
            count (int): total number of connections to add

        Returns:
            matrix (torch.Tensor): the updated matrix.
        """
        n_nodes = array.shape[0]

        src = np.random.choice(n_nodes, count)
        dest = np.random.choice(n_nodes, count)
        polarities = self.polarity((count,))

        array[src, dest] = polarities
        return array

    def build(self) -> None:
        """
        Builds the mask wiring for each layer.

        !!! note "Layer format"

            Follows a three layer format, each with separate masks:

            1. Sensory -> inter
            2. Inter -> command
            3. Command -> motor

        Plus, command recurrent connections for ODE solvers.
        """
        # Sensory -> inter
        self.masks.inter = self._build_connections(
            self.masks.inter,
            self._n_connections.sensory,
        )
        # Inter -> command
        self.masks.command = self._build_connections(
            self.masks.command,
            self._n_connections.inter,
        )
        # Command -> motor
        self.masks.motor = self._build_connections(
            self.masks.motor,
            self._n_connections.command,
        )

        # Command -> command
        self.masks.recurrent = self._build_recurrent_connections(
            self.masks.recurrent,
            self._n_connections.command,
        )

    def data(self) -> Tuple[LayerMasks, NeuronCounts]:
        """
        Retrieves wiring storage containers for layer masks and node counts.

        Returns:
            masks (LayerMasks): the object containing layer masks.
            counts (NeuronCounts): the object containing node counts.
        """
        return self.masks, self.counts



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal

import gymnasium as gym
from sqlmodel import Session

from velora.models.base import RLModuleAgent

StatusLiteral = Literal["start", "episode", "logging", "step", "complete"]
RecordMethodLiteral = Literal["episode", "step"]


@dataclass
class RecordState:
    """
    A storage container for the video recording state.

    Parameters:
        dirpath (Path): the video directory path to store the videos
        method (Literal["episode", "step"]): the recording method
        episode_trigger (Callable[[int], bool], optional): the `episode` recording
            trigger
        step_trigger (Callable[[int], bool], optional): the `step` recording trigger
    """

    dirpath: Path
    method: RecordMethodLiteral
    episode_trigger: Callable[[int], bool] | None = None
    step_trigger: Callable[[int], bool] | None = None

    def to_wrapper(self) -> Dict[str, Any]:
        """
        Converts the state into wrapper parameters.

        Returns:
            params (Dict[str, Any]): values as parameters for [Gymnasium's RecordVideo](https://gymnasium.farama.org/api/wrappers/misc_wrappers/#gymnasium.wrappers.RecordVideo) wrapper.

            Includes the following keys - `[video_folder, episode_trigger, step_trigger]`.
        """
        return {
            "video_folder": self.dirpath,
            "episode_trigger": self.episode_trigger,
            "step_trigger": self.step_trigger,
        }


@dataclass
class AnalyticsState:
    """
    A storage container for the details of a [Comet](https://www.comet.com/) or
    [Weights and Biases](https://wandb.ai/) analytics experiment.

    Parameters:
        project_name (str): the name of the project to add this experiment to
        experiment_name (str, optional): the name of the experiment
        tags (List[str], optional): a list of tags associated with the experiment
    """

    project_name: str
    experiment_name: str | None = None
    tags: List[str] | None = None


@dataclass
class TrainState:
    """
    A storage container for the current state of model training.

    Parameters:
        agent (RLModuleAgent): the agent being trained
        env (gymnasium.Env): a single training or evaluation environment
        session (sqlmodel.Session): the current metric database session
        experiment_id (int): the current experiment's unique ID
        total_episodes (int, optional): total number of training episodes
        total_steps (int, optional): total number of training steps
        status (Literal["start", "episode", "logging", "step", "complete"], optional): the current stage of training.

            - `start` - before training starts.
            - `episode` - inside the episode loop.
            - `logging` - metric logging.
            - `step` - inside the timestep loop.
            - `complete` - completed training.

        logging_type (Literal["episode", "step"], optional): the logging type
        current_ep (int, optional): the current episode index
        current_step (int, optional): the current training timestep
        ep_reward (float, optional): the current episode reward
        stop_training (bool, optional): a flag to declare training termination
        saving_enabled (bool, optional): a flag for checkpoint saving
        checkpoint_dir (Path, optional): the checkpoint directory path when
            `saving_enabled=True`
        record_state (RecordState, optional): the video recording state
        analytics_state (AnalyticsState, optional): the analytics state
    """

    agent: RLModuleAgent
    env: gym.Env
    session: Session
    experiment_id: int
    total_episodes: int = 0
    total_steps: int = 0
    status: StatusLiteral = "start"
    logging_type: Literal["episode", "step"] = "episode"
    current_ep: int = 0
    current_step: int = 0
    ep_reward: float = 0.0
    stop_training: bool = False
    saving_enabled: bool = False
    checkpoint_dir: Path | None = None
    record_state: RecordState | None = None
    analytics_state: AnalyticsState | None = None

    def update(
        self,
        *,
        status: StatusLiteral | None = None,
        current_ep: int | None = None,
        current_step: int | None = None,
        ep_reward: int | None = None,
        logging_type: Literal["episode", "step"] | None = None,
    ) -> None:
        """
        Updates the training state. When any input is `None`, uses existing value.

        Parameters:
            status (Literal["start", "episode", "logging", "step", "complete"], optional): the current stage of training.

                - `start` - before training start.
                - `episode` - inside the episode loop.
                - `logging` - metric logging.
                - `step` - inside the timestep loop.
                - `complete` - completed training.

            current_ep (int, optional): the current episode index
            current_step (int, optional): the current training timestep
            ep_reward (float, optional): the current episode or rollout update reward
            logging_type (Literal["episode", "step"], optional): the logging type
        """
        self.status = status if status else self.status
        self.current_ep = current_ep if current_ep else self.current_ep
        self.current_step = current_step if current_step else self.current_step
        self.ep_reward = ep_reward if ep_reward else self.ep_reward
        self.logging_type = logging_type if logging_type else self.logging_type

    def analytics_update(self) -> None:
        """
        Updates the analytics state details that are `None` dynamically, using
        the current training state.
        """
        agent_name = self.agent.__class__.__name__
        env_name = self.env.spec.name

        new_state = self.analytics_state

        new_state.experiment_name = (
            new_state.experiment_name
            if new_state.experiment_name
            else f"{agent_name}_{env_name}_{self.total_episodes}ep"
        )

        new_state.tags = new_state.tags if new_state.tags else [agent_name, env_name]

        # Update state
        self.analytics_state = new_state



from typing import Any, Dict, List

import torch
import torch.nn as nn


def to_tensor(
    items: List[Any],
    *,
    dtype: torch.dtype = torch.float32,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Converts a list of items to a Tensor, then:

    1. Converts it to a specific `dtype`
    2. Loads it onto `device`

    Parameters:
        items (List[Any]): a list of items of any type
        dtype (torch.dtype, optional): the data type for the tensor
        device (torch.device, optional): the device to perform computations on

    Returns:
        tensor (torch.Tensor): the updated `torch.Tensor`.
    """
    return torch.tensor(items).to(dtype=dtype, device=device)


def stack_tensor(
    items: List[torch.Tensor],
    *,
    dtype: torch.dtype = torch.float32,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Stacks a list of tensors together, then:

    1. Converts it to a specific `dtype`
    2. Loads it onto `device`

    Parameters:
        items (List[torch.Tensor]): a list of torch.Tensors full of items
        dtype (torch.dtype, optional): the data type for the tensor
        device (torch.device, optional): the device to perform computations on

    Returns:
        tensor (torch.Tensor): the updated `torch.Tensor`.
    """
    return torch.stack(items).to(dtype=dtype, device=device)


def soft_update(source: nn.Module, target: nn.Module, *, tau: float = 0.005) -> None:
    """
    Performs a soft parameter update between two PyTorch Networks.

    Parameters:
        source (nn.Module): the source network
        target (nn.Module): the target network
        tau (float, optional): the soft update factor used to slowly update
            the target network
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(tau * param.data + (1.0 - tau) * target_param.data)


def hard_update(source: nn.Module, target: nn.Module) -> None:
    """
    Performs a hard parameter update between two PyTorch Networks.

    Parameters:
        source (nn.Module): the source network
        target (nn.Module): the target network
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(param.data)


@torch.jit.ignore
def total_parameters(model: nn.Module) -> int:
    """
    Calculates the total number of parameters used in a PyTorch `nn.Module`.

    Parameters:
        model (nn.Module): a PyTorch module with parameters

    Returns:
        count (int): the total number of parameters.
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


@torch.jit.ignore
def active_parameters(model: nn.Module) -> int:
    """
    Calculates the active number of parameters used in a PyTorch `nn.Module`.
    Filters out parameters that are `0`.

    Parameters:
        model (nn.Module): a PyTorch module with parameters

    Returns:
        count (int): the total active number of parameters.
    """
    return sum((p != 0).sum().item() for p in model.parameters() if p.requires_grad)


@torch.jit.ignore
def summary(module: nn.Module) -> Dict[str, str]:
    """
    Outputs a summary of a module and all it's sub-modules as a dictionary.

    Returns:
        summary (Dict[str, str]): key-value pairs for the network layout.
    """
    model_dict = {}

    for name, mod in module.named_children():
        if len(list(mod.children())) > 0:
            # If the module has submodules, recurse
            model_dict[name] = summary(mod)
        else:
            # If it's a leaf module, store its string representation
            model_dict[name] = str(mod)

    return model_dict



from datetime import datetime
from typing import List

from sqlmodel import Field, Relationship, SQLModel


class Experiment(SQLModel, table=True):
    """
    Experiment information tracking agent, environment, and metadata.

    Attributes:
        id (int): unique identifier for the experiment
        agent (str): the name of the agent used in the experiment
        env (str): the name of the environment used in the experiment
        config (str): a JSON string containing the agent's configuration details
        created_at (datetime): the date and time the experiment was created
    """

    id: int | None = Field(default=None, primary_key=True)
    agent: str = Field(index=True)
    env: str = Field(index=True)
    config: str  # JSON object
    created_at: datetime = Field(default_factory=datetime.now)

    # Relationships
    episodes: List["Episode"] = Relationship(back_populates="experiment")


class Episode(SQLModel, table=True):
    """
    Episode-level metrics tracking reward, length, and agent performance.

    Attributes:
        id (int): a unique identifier for the episode
        experiment_id (int): the experiment ID associated to the episode
        episode_num (int): the episode index
        reward (float): the episodic reward (return)
        length (int): the number of timesteps performed to terminate the episode
        reward_moving_avg (float): the episodes reward moving average based on a window size
        reward_moving_std (float): the episodes reward moving standard deviation based on a
            window size
        actor_loss (float): the average Actor loss for the episode
        critic_loss (float): the average Critic loss for the episode
        entropy_loss (float): the average Entropy loss for the episode
        created_at (datetime): the date and time when the the entry was created
    """

    id: int | None = Field(default=None, primary_key=True)
    experiment_id: int = Field(foreign_key="experiment.id", index=True)
    episode_num: int = Field(index=True)

    # Core metrics
    reward: float
    length: int

    # Statistical metrics
    reward_moving_avg: float
    reward_moving_std: float

    # Loss metrics
    actor_loss: float
    critic_loss: float
    entropy_loss: float = Field(default=0.0)

    # Timestamps
    created_at: datetime = Field(default_factory=datetime.now)

    # Relationships
    experiment: Experiment = Relationship(back_populates="episodes")



import time
from dataclasses import dataclass
from typing import Self


@dataclass
class ElapsedTime:
    """
    A storage container for time tracking.

    Parameters:
        hrs (float) hours taken
        mins (float) minutes taken
        secs (float) seconds taken
    """

    hrs: float
    mins: float
    secs: float

    @classmethod
    def elapsed(cls, start_time: float) -> Self:
        """
        Calculates the elapsed time from `now` and a `start_time`.

        Parameters:
            start_time (float): the start time of an event

        Returns:
            self (Self): a newly populated storage container.
        """
        elapsed = time.time() - start_time
        hrs, remainder = divmod(elapsed, 3600)
        mins, secs = divmod(remainder, 60)

        return cls(hrs=hrs, mins=mins, secs=secs)

    def __str__(self) -> str:
        return f"{self.hrs:.2f}h {self.mins:.2f}m {self.secs:.2f}s"
